/*
Robots.h
Written by Walter O. Krawec

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

// contains code to support a simulated robot

#ifndef SRS10_ROBOTS_
#define SRS10_ROBOTS_
//#include "sockets/sockets.h"
#include "../wok/Sockets.h"
#include "../wok/Timer.h"
#include <list>

#include "Shapes.h"
#include "Track.h"
#include "Entity.h"

#include "../wok/UID.h"
class Overlay
{
public:
	struct Point
	{
		Point() : x(0), y(0), r(0), g(0), b(0) {};
		Point(float _x, float _y) : x(_x), y(_y), r(0), g(0), b(0) {};
		Point(float _x, float _y, float _r, float _g, float _b) : x(_x), y(_y), r(_r), g(_g), b(_b) {};
		float x, y;
		float r, g, b;
	};
	struct TwoPoints
	{
		TwoPoints() : x1(0), y1(0), x2(0), y2(0) {};
		TwoPoints(Point& p1, Point& p2)
		{
			x1 = p1.x;
			y1 = p1.y;

			x2 = p2.x;
			y2 = p2.y;
		}

		float x1, y1, x2, y2;
	};

	void addSquare(Point& p) {Squares.push_back(p);};
	void addLine(TwoPoints& p){Lines.push_back(p);};
	void clearAll()
	{
		Squares.clear();
		Lines.clear();
	}

	void clearSquares()
	{
		Squares.clear();
	}

	void draw()
	{
		if(!Squares.empty())
		{
			float size = 0.1f;
			std::list <Point>::iterator Iter;
			for(Iter = Squares.begin(); Iter != Squares.end(); ++Iter)
			{
				glColor4f(Iter->r, Iter->g, Iter->b, 1);
				glBegin(GL_QUADS);
					glVertex3f(Iter->x, Iter->y, 0);
					glVertex3f(Iter->x+size, Iter->y, 0);
					glVertex3f(Iter->x+size, Iter->y+size, 0);
					glVertex3f(Iter->x, Iter->y+size, 0);
				glEnd();
			}
		}
		glColor4f(0,0,1,1);
		if(!Lines.empty())
		{
			std::list <TwoPoints>::iterator Iter;
			for(Iter = Lines.begin(); Iter != Lines.end(); ++Iter)
			{
				glLineWidth(4.0f);
				glBegin(GL_LINES);
					glVertex3f(Iter->x1, Iter->y1, 0);
					glVertex3f(Iter->x2, Iter->y2, 0);
				glEnd();
				glLineWidth(1.0f);
			}
		}

		glColor4f(1,1,1,1);
	}
private:
	std::list <Point> Squares;
	std::list <TwoPoints> Lines;
};

namespace SRS10{
namespace robot
{
	enum Direction
	{
		Forward=0,
		Left,
		Right,
		Back,
		Stop
	};
	
	struct SensorData
	{
		wok::UID ID;
		int value;
	};
	
	struct Event
	{
		Event() : x(0), y(0), type(0), lastMoveCommand(0), lastSpecialMoveCommand(0), creator(0), direction(0) {};
		Event(float _x, float _y, int _type, int _move, char _creator) : x(_x), y(_y), type(_type), lastMoveCommand(_move), creator(_creator) {};
		
		//std::vector <int> inputs, outputs;
		std::vector <SensorData> sensors;
		float x, y;
		int lastMoveCommand, lastSpecialMoveCommand;
		int type;
		char creator;
		int direction;	// a value 0-359 that specifies degree rotation of event location
	};
	
	class RobotManager;
	
	class Sensor
	{
	public:
		Sensor() : x(0), y(0), rot(0) {};
		Sensor(float _x, float _y, float _rot) : x(_x), y(_y), rot(_rot) {};

		virtual int get(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID)=0;
		virtual int getLast() { return last; };
		virtual void draw(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID)=0;
		
		virtual void set(float arg1, float arg2) {};
		
		wok::UID getID() {return ID;}

	protected:
		float x,y,rot;
		int last;
		
		wok::UID ID;
	};

	class LightSensor : public Sensor
	{
	public:
		LightSensor() {};
		LightSensor(float distance, float _rot) : Sensor(distance, 0, _rot) {};

		virtual void draw(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID);
		virtual int get(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID);
	};

	class LineSensor : public Sensor
	{
	public:
		LineSensor() : cooldown(0) {};
		LineSensor(float distance, float _rot) : Sensor(distance, 0, _rot), cooldown(0) {};

		virtual void draw(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID);
		virtual int get(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID);

	private:
		int cooldown;
	};

	class ProximitySensor : public Sensor
	{
	public:
		ProximitySensor() : range(0), quad(0) {};
		ProximitySensor(float _range, int _quad) : range(_range), quad(_quad) {};

		virtual void draw(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID);
		virtual int get(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID);

	private:
		float range;
		//int theta1, theta2;
		int quad;
	};

	class IR : public Sensor
	{
	public:
		IR() : range(0.75f) {};
		IR(float _x, float _y, float _rot, float _range) : Sensor(_x, _y, _rot), range(_range) {};

		virtual void draw(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID);

		virtual int get(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID);

	private:
		float range;
	};

	class IR360 : public Sensor
	{
	public:
		IR360() : range(0.75f), autoScan(false), destinationRot(0), lastDistance(0) {};
		IR360(float _x, float _y, float _rot, float _range, bool _autoScan) : Sensor(_x, _y, _rot), lastDistance(0), range(_range), destinationRot(_rot), autoScan(_autoScan) {if(autoScan){destinationRot=90;}};
		
		virtual void draw(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID);
		virtual int get(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID);
		virtual void set(float arg1, float arg2); // arg1 = destinationRot

		bool isRotation(float queryRot)
		{
			if(fabs(rot - queryRot) < 0.01)
				return true;
			return false;
		}

		bool isRotating()
		{
			return !isRotation(destinationRot);
		}
		float getRotation() {return rot;};
		float getRange() {return range;};
		float getLastDistance() {return lastDistance;};
		
	private:
		void updateRotation()
		{
			float dr = SRS10::graphics::GraphicsEngine::instance().getSimulationSpeed()*205.0f/SRS10::graphics::GraphicsEngine::instance().getFPS();
			if(fabs(rot-destinationRot) < dr)
			{
				rot = destinationRot;
				if(autoScan)
					destinationRot *= -1;
			}
			else
			{
				if(rot < destinationRot)
					rot += dr;
				else
					rot -= dr;
			}
		}
		
		float range, destinationRot;
		bool autoScan;
		float lastDistance;
	};
	
	class Timer : public Sensor
		{
		public:
			Timer() : counter(0) {};
			
			virtual void draw(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID)
			{
				counter += (50.0f/SRS10::graphics::GraphicsEngine::instance().getFPS());
				if(counter > 360)
					counter = 0;
			}
			
			virtual int get(float dx, float dy, float dr, SRS10::track::Track& track, RobotManager& manager, int robotID)
			{
				return 1;	// sensors used to be floats...
				//return (sin(counter*(3.14f/180)));
			}
			
		private:
			float counter;
		};

	class SensorManager
		{
		public:
			SensorManager() {};
			~SensorManager() {cleanup();};
			
			void cleanup()
			{
				for(int i=0; i<(int)sensors.size(); ++i)
				{
					if(sensors[i])
						delete sensors[i];
				}
				
				sensors.clear();
			}
			
			wok::UID add(Sensor* s)
			{
				if(!s)
					return wok::UID();
				
				sensors.push_back(s);
				
				return s->getID();
			}
			
			Sensor* get(wok::UID& ID)
			{
				for(int i=0; i<(int)sensors.size(); ++i)
				{
					if(sensors[i] && sensors[i]->getID() == ID)
						return sensors[i];
				}
				
				return NULL;
			}
			
			Sensor* get(int ID)
			{
				wok::UID uID;
				uID.set(ID);
				return get(uID);
			}
			
			std::vector <Sensor*>& getSensorVector() {return sensors;};
			
		private:
			std::vector <Sensor*> sensors;
		};

	class Robot;

	class ANP
	{
		enum ANPData
		{
			Blocked=0,
			Warning=1,
			Clear=2
		};

	public:
		ANP()
		{
			runningANP = 0;
			rotationANP = 0;
			Overlay::Point lastPointANP;
			lastPointIsValidANP=false;
			precisionANP = 4;
			scanCounterANP = 0;

			linePrecision = 20;

			dataANP = new ANPData[2*precisionANP+1];

			navPoint = -1;

			reportCollision = false;
		}
		~ANP()
		{
			if(dataANP)
				delete[] dataANP;
			dataANP = NULL;
		}

		void runANP(SRS10::robot::Robot* robot, SRS10::robot::IR360* sensorANP, Overlay* hudANP);
		bool navToPoint(SRS10::robot::Robot* robot, int point, Overlay* hudANP);
		void start(SRS10::robot::Robot* robot);

		bool isScanning()
		{
			if(runningANP >= 1 && runningANP <= 4)
				return true;
			return false;
		}
		bool isMoving()
		{
			if(runningANP >= 5)
				return true;
			return false;
		}

		int getMaxNavPoints() {return 2*precisionANP+1;};
		ANPData getNavPoint(int point)
		{
			if(dataANP && point >= 0 && point < 2*precisionANP+1)
				return dataANP[point];
			return Blocked;
		}

		bool collision() {return reportCollision;}

	private:
		ANPData *dataANP;
		int runningANP;
		double rotationANP;
		Overlay::Point lastPointANP;
		bool lastPointIsValidANP;
		int precisionANP;
		int scanCounterANP;
		int linePrecision;

		int navPoint;
		math3D::CVector2D startPoint;
		float startRot;
		bool reportCollision;
	};
	
	class Robot
	{
	public:
		friend class RobotManager;

		Robot() : useANP(false), selected(false), x(0), y(0), turnSpeed(0), speed(0), rot(0), sphere(0.25f), dir(Stop), ID(0), bump(false), robotBump(false), lastMoveCommand(0), lastSpecialMoveCommand(0), type(0) {};
		~Robot() {cleanup();}

		void cleanup()
		{
			shapeList.cleanup();
			/*for(size_t i=0; i<sensors.size(); ++i)
			{
				if(sensors[i])
					delete sensors[i];
			}

			sensors.clear();*/
		}

		bool loadFromFile(const std::string& filename);
		void update(track::Track& track, RobotManager& manager, bool selected=false);

		void setDirection(Direction dir)
		{
			Direction lastDir = this->dir;
			this->dir = dir;
		}

		//void runAI(track::Track& track, RobotManager& manager, bool bump, bool robotBump);

		void setRotation(float r) {rot = r;}

		void setPosition(float _x, float _y) {x=_x; y=_y; lastPos.x = x; lastPos.y = y; lastPos.z = 0;}
		
		SensorManager& getSensorManager() {return sensorManager;};
		
		void packSensorValues(wok::sockets::Packet& p, SRS10::track::Track& track, RobotManager& manager);
		void packEvents(wok::sockets::Packet& p, SRS10::track::Track& track, RobotManager& manager);
		
		void clearEvents() {events.clear();};

		void addEvent(float eventX, float eventY, int eventR, char creator);
		
		void runMoveCommand(int cmd, int specialCmd)
		{
			switch(cmd) {
				case 0: speed = 0; turnSpeed = 0; break;
				case 1: speed = 1.0f; turnSpeed = 0; break;
				case 2: speed = -1.0f; turnSpeed = 0; break;
				case 3: speed = 0; turnSpeed = 1.0f; break;
				case 4: speed = 0; turnSpeed = -1.0f; break;
			}

			lastMoveCommand = cmd;
			lastSpecialMoveCommand = specialCmd;
		}
		
		float getX() {return x;}
		float getY() {return y;}
		float getRot() {return rot;}

		void startANP() {anp.start(this);}
		void runANP() {anp.runANP(this, &anpIR, &hudANP);}
		void anpNavToPoint(int point) {anp.navToPoint(this, point, &hudANP);}

	//private:
		ANP anp;
		IR360 anpIR;
		bool useANP;
		Overlay hudANP;
		SensorManager sensorManager;
		SRS10::graphics::ShapeCollection shapeList;
		std::list <Event> events;
		//std::vector <Sensor*> sensors;
		
		float x,y,rot;	// position, rotation (deg.)
		float turnSpeed;
		float sphere;	// for collision
		Direction dir;
		float speed;

		int lastMoveCommand, lastSpecialMoveCommand;

		math3D::CVector3D lastPos;
		
		int ID;
		
		bool bump, robotBump;

		char type;

		std::vector <int> inputVector, outputVector;

		bool selected;
	};

	
	class RobotManager
	{
	public:
		RobotManager() : nextID(0), selectedRobot(-1) {}
		~RobotManager() {cleanup();}

		void cleanup()
		{
			for(int i=0; i<(int)robots.size(); ++i)
			{
				if(robots[i])
					delete robots[i];
			}

			robots.clear();
		}

		int add(Robot* r)
		{
			if(!r)
				return -1;

			r->ID = nextID;
			++nextID;
			robots.push_back(r);
			
			return r->ID;
		}
		
		Robot* getRobot(int ID)
		{
			for(int i=0; i<(int)robots.size(); ++i)
			{
				if(robots[i] && robots[i]->ID == ID)
					return robots[i];
			}
			
			return NULL;
		}
		
		void rotateAll180()
		{
			for(int i=0; i<(int)robots.size(); ++i)
			{
				if(robots[i])
					robots[i]->rot += 180;
			}
		}

		void updateAll(track::Track& track)
		{
			for(int i=0; i<(int)robots.size(); ++i)
			{
				if(robots[i])
				{
					if(selectedRobot == robots[i]->ID)
						robots[i]->update(track, *this, true);
					else
						robots[i]->update(track, *this, false);
				}
			}
		}

		bool collision(SRS10::graphics::CollisionCircle& circle, int ID);
		
		void selectRobot(int ID)
		{
			selectedRobot = ID;
		}

		void addGlobalEvent(float x, float y, int r, char creator, float volume, int ignoreID=-1)
		{
			for(unsigned int i=0; i<robots.size(); ++i)
			{
				//if(ignoreID >= 0 && i == (int)ignoreID) continue;
				if(!robots[i]) continue;
				if(ignoreID >= 0 && robots[i]->ID == ignoreID) continue;

				if(volume < 0 || (robots[i]->x-x)*(robots[i]->x-x) + (robots[i]->y-y)*(robots[i]->y-y) <= volume)
					robots[i]->addEvent(x,y,r, creator);
			}
		}

		std::vector <Robot*>& getRobotList() {return robots;};
		
	private:
		std::vector <Robot*> robots;
		int nextID;
		
		int selectedRobot;
	};
}
}
#endif