/*
SRS10Client.cpp
Written by Walter O. Krawec

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*
Include this file to access SRS10 in your C++ program;
Alternatively use the SRS10Module in wok::Interpreter
*/

#ifndef SRS10_CLIENT_H_
#define SRS10_CLIENT_H_

#include <WOK/Sockets.h>
#include <WOK/UID.h>

// Change this if you make any major changes to the packet structure
// to avoid conflicts between versions
#define PACKET_VERSION 2

namespace SRS10
{	
	struct SensorData
	{
		SensorData() : value(0) {};
		SensorData(const std::string& _name, int _ID) : name(_name), value(0)
		{
			ID.set(_ID);
		};
		
		int getValue() {return value;};
		void setValue(int val) {value = val;};
		
		std::string name;
		wok::UID ID;
		
		int value;
	};

	struct Event
	{
		Event() : x(0), y(0), type(0), lastMoveCommand(0), creator(0), direction(0) {};

		float x, y;
		int type, lastMoveCommand, lastSpecialMoveCommand;
		char creator;
		std::vector <SensorData> sensors;

		int direction;	// the angle (0-359) of the event's location
	};
	
	struct RobotData
	{
		RobotData() : trackCollision(false), robotCollision(false), x(0), y(0), lastX(0), lastY(0), rotation(0) {};
		RobotData(const std::string& _name, int _ID) : name(_name), trackCollision(false), robotCollision(false), x(0), y(0), lastX(0), lastY(0)
		{
			ID.set(_ID);
			rotation = 0;
		};
		
		std::vector <SensorData> sensors;
		std::list <Event> events;
		std::string name;
		wok::UID ID;
		
		bool trackCollision, robotCollision;
		float x,y, lastX, lastY;
		float rotation;
	};
	
	class Client
	{
	public:
		enum Direction
		{
			Forward=0,
			Left,
			Right,
			Back
		};
		bool connectToHost(const std::string& IPAddr, int portNumber)
		{
			//client.getSocket()->initWinSock();
			return client.connectToHost(IPAddr.c_str(), portNumber);
		}
		void disconnect()
		{
			client.disconnect();
			robots.clear();
		}
		float getEventSensorValue(const std::string& sensorName, Event& e)
		{
			for(int i=0; i<(int)e.sensors.size(); ++i)
			{
				if(e.sensors[i].name == sensorName)
					return (float)e.sensors[i].getValue();
			}
			return 0;
		}
		Direction getEventDirection(Event& e)
		{
			int t = e.direction;
			if(t < 45 || t > 315)
				return Forward;
			else if(t >= 45 && t < 135)
				return Left;
			else if(t >= 135 && t < 225)
				return Back;

			return Right;
		}
		bool getNextEvent(const std::string& robotName, Event& e)
		{
			int r = 0;
			if(!getRobotData(robotName, r))
				return false;

			if(robots[r].events.empty())
				return false;

			e = *robots[r].events.begin();
			robots[r].events.erase(robots[r].events.begin());
			return true;
		}
		bool eventsInList(const std::string& robotName)
		{
			int r = 0;
			if(!getRobotData(robotName, r))
				return false;

			if(robots[r].events.empty())
				return false;

			return true;
		}
		bool getRobotData(const std::string& name, int& robotIndex)
		{
			for(unsigned int i=0; i<robots.size(); ++i)
			{
				if(robots[i].name == name)
				{
					robotIndex = i;
					return true;
				}
			}
			
			return false;
		}
		
		bool getSensorData(const std::string& robotName, const std::string& sensorName, int& robotIndex, int& sensorIndex)
		{
			if(!getRobotData(robotName, robotIndex))
				return false;
			
			for(unsigned int i=0; i<robots[robotIndex].sensors.size(); ++i)
			{
				if(robots[robotIndex].sensors[i].name == sensorName)
				{
					sensorIndex = i;
					return true;
				}
			}
			
			return false;
		}
		bool getSensorData(const std::string& robotName, int ID, int& robotIndex, int& sensorIndex)
		{
			if(!getRobotData(robotName, robotIndex))
				return false;
			
			for(unsigned int i=0; i<robots[robotIndex].sensors.size(); ++i)
			{
				if(robots[robotIndex].sensors[i].ID.get() == ID)
				{
					sensorIndex = i;
					return true;
				}
			}
			
			return false;
		}
		bool addRobot(const std::string& name, char type=0)
		{
			wok::sockets::Packet p;
			p.createHeader(1);
			p.packChar(type);
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);
			if(!waitForResponse(p, 1))
				return false;
			
			robots.push_back(RobotData(name, p.unpackInt()));
			
			return true;
		}
		
		bool setIOVector(const std::string& robotName, const std::vector <int> inputs, const std::vector <int> outputs)
		{
			int robotIndex = 0;
			if(!getRobotData(robotName, robotIndex))
				return false;

			wok::sockets::Packet p;
			p.createHeader(13);
			for(unsigned int i=0; i<inputs.size(); ++i)
				p.packChar((char)inputs[i]);
			p.packInt((int)inputs.size());

			for(unsigned int i=0; i<outputs.size(); ++i)
				p.packChar((char)outputs[i]);
			p.packInt((int)outputs.size());

			p.packInt(robots[robotIndex].ID.get());
			p.packChar(PACKET_VERSION);
			return true;
		}

		// ANP was a test navigation program
		bool useANP(const std::string& robotName)
		{
			int robotIndex=0;
			if(!getRobotData(robotName, robotIndex))
				return false;

			wok::sockets::Packet p;
			p.createHeader(50);
			p.packInt(robots[robotIndex].ID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);
			return true;
		}
		bool startANP(const std::string& robotName)
		{
			std::cout << "start anp\n";
			int robotIndex=0;
			if(!getRobotData(robotName, robotIndex))
				return false;

			wok::sockets::Packet p;
			p.createHeader(51);
			p.packInt(robots[robotIndex].ID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);
		}
		bool getANPData(const std::string& robotName, std::vector <int>& anpData)
		{
			//std::cout << "get data\n";
			int robotIndex=0;
			if(!getRobotData(robotName, robotIndex))
				return false;

			wok::sockets::Packet p;
			p.createHeader(52);
			p.packInt(robots[robotIndex].ID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);

			if(!waitForResponse(p, 52))
				return false;

			if(p.unpackChar() == 0)
				return false;

			int N = p.unpackInt();
			//std::cout << N << "\n";
			for(int i=0; i<N; ++i)
				anpData.push_back(p.unpackInt());

			return true;
		}
		bool anp_navToPoint(const std::string& robotName, int point)
		{
		//	std::cout << "nav to point\n";
			int robotIndex=0;
			if(!getRobotData(robotName, robotIndex))
				return false;

			wok::sockets::Packet p;
			p.createHeader(53);
			p.packInt(point);
			p.packInt(robots[robotIndex].ID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);
		}

		bool anp_getStatus(const std::string& robotName, bool& collision)	// returns false if busy
		{
			//std::cout << "get status\n";
			int robotIndex=0;
			if(!getRobotData(robotName, robotIndex))
				return false;

			wok::sockets::Packet p;
			p.createHeader(54);
			p.packInt(robots[robotIndex].ID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);

			if(!waitForResponse(p, 54))
				return false;

			char ready = p.unpackChar();
			char c = p.unpackChar();
			if(c == 1)
				collision = true;
			else
				collision = false;
			if(ready == 1)
				return true;
			return false;
		}
		bool addLightSensor(const std::string& robotName, const std::string& sensorName, float x, float y)
		{
			int robotIndex=0;
			if(!getRobotData(robotName, robotIndex))
				return false;
			
			wok::sockets::Packet p;
			p.createHeader(2);
			p.packFloat(x);
			p.packFloat(y);
			p.packChar(3);
			p.packInt(robots[robotIndex].ID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);
			if(!waitForResponse(p, 2))
				return false;
			
			robots[robotIndex].sensors.push_back(SensorData(sensorName, p.unpackInt()));
			return true;
		}
		bool addLineSensor(const std::string& robotName, const std::string& sensorName, float x, float y)
		{
			int robotIndex=0;
			if(!getRobotData(robotName, robotIndex))
				return false;
			
			wok::sockets::Packet p;
			p.createHeader(2);
			p.packFloat(x);
			p.packFloat(y);
			p.packChar(5);
			p.packInt(robots[robotIndex].ID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);
			if(!waitForResponse(p, 2))
				return false;
			
			robots[robotIndex].sensors.push_back(SensorData(sensorName, p.unpackInt()));
			return true;
		}
		bool addRobotProximitySensor(const std::string& robotName, const std::string& sensorName, float range, int quad)
		{
			int robotIndex=0;
			if(!getRobotData(robotName, robotIndex))
				return false;
			
			wok::sockets::Packet p;
			p.createHeader(2);
			p.packFloat(range);
			p.packInt(quad);
			p.packChar(4);
			p.packInt(robots[robotIndex].ID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);
			if(!waitForResponse(p, 2))
				return false;
			
			robots[robotIndex].sensors.push_back(SensorData(sensorName, p.unpackInt()));
			return true;
		}
		bool addIRSensor(const std::string& robotName, const std::string& sensorName, float range, float rotation)
		{
			int robotIndex=0;
			if(!getRobotData(robotName, robotIndex))
				return false;
			
			wok::sockets::Packet p;
			p.createHeader(2);
			p.packFloat(rotation);
			p.packFloat(range);
			p.packChar(1);
			p.packInt(robots[robotIndex].ID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);
			if(!waitForResponse(p, 2))
				return false;
			
			robots[robotIndex].sensors.push_back(SensorData(sensorName, p.unpackInt()));
			return true;
		}
		
		bool addIR360Sensor(const std::string& robotName, const std::string& sensorName, float range, float rotation, bool autoScan)
		{
			int robotIndex=0;
			if(!getRobotData(robotName, robotIndex))
				return false;
			
			wok::sockets::Packet p;
			p.createHeader(2);
			if(autoScan)
				p.packChar(1);
			else
				p.packChar(0);
			p.packFloat(rotation);
			p.packFloat(range);
			p.packChar(2);
			p.packInt(robots[robotIndex].ID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);
			if(!waitForResponse(p, 2))
				return false;
			
			robots[robotIndex].sensors.push_back(SensorData(sensorName, p.unpackInt()));
			return true;
		}

		bool move(const std::string& robotName, float forwardSpeed, float turnSpeed)
		{
			int r=0;
			if(!getRobotData(robotName, r))
				return false;

			wok::sockets::Packet p;
			p.createHeader(15);
			p.packFloat(forwardSpeed);
			p.packFloat(turnSpeed);

			p.packInt(robots[r].ID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);
			return true;
		}
		
		// cmd:
		// 0: stop
		// 1: Forward
		// 2: Backward
		// 3: Left?
		// 4: Right?
		bool moveCommand(const std::string& robotName, int cmd, int specialCmd=0)
		{
			int r=0;
			if(!getRobotData(robotName, r))
				return false;
			
			wok::sockets::Packet p;
			p.createHeader(10);
			p.packChar(cmd);
			p.packChar(specialCmd);
			p.packInt(robots[r].ID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);
			return true;
		}
		
		bool sensorCommand(const std::string& robotName, const std::string& sensorName, float arg1, float arg2)
		{
			int r=0, s=0;
			if(!getSensorData(robotName, sensorName, r, s))
				return false;
			
			wok::sockets::Packet p;
			p.createHeader(12);
			p.packFloat(arg1);
			p.packFloat(arg2);
			p.packInt(robots[r].sensors[s].ID.get());
			p.packInt(robots[r].ID.get());
			
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);
			return true;
		}
		
		bool createEvent(const std::string& robotName, float volume, char creator, int r_val, float red, float green, float blue)
		{
			int r = 0;
			if(!getRobotData(robotName, r))
				return false;

			wok::sockets::Packet p;
			p.createHeader(14);

			p.packFloat(red);
			p.packFloat(green);
			p.packFloat(blue);

			p.packInt(r_val);
			p.packFloat(volume);
			p.packChar(creator);

			p.packInt(robots[r].ID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);

			return true;
		}

		bool pollEvents(const std::string& robotName)
		{
			int r = 0;
			if(!getRobotData(robotName, r))
				return false;

			wok::UID callID;
			wok::sockets::Packet p;
			p.createHeader(21);
			p.packInt(robots[r].ID.get());
			p.packInt(callID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);

			if(!waitForResponse(p, 21))
				return false;

			if(p.unpackInt() != callID.get())
				return false;

			int N = p.unpackInt();
			SensorData data;
			Event e;
			for(int i=0; i<N; ++i)
			{
				int M = p.unpackChar();
				e.sensors.clear();
				for(int j=0; j<M; ++j)
				{
					int v = p.unpackInt();
					int sensorID = p.unpackInt();
					data.ID.set(sensorID);
					data.value = v;
					for(unsigned int k=0; k<robots[r].sensors.size(); ++k)
					{
						if(robots[r].sensors[k].ID == data.ID)
							data.name = robots[r].sensors[k].name;
					}
					e.sensors.push_back(data);
				}
				int direction = p.unpackInt();
				char creator = p.unpackChar();
				int special = p.unpackInt();
				int move = p.unpackInt();
				int type = p.unpackInt();
				float y = p.unpackFloat();
				float x = p.unpackFloat();
				e.type = type;
				e.x = x;
				e.y = y;
				e.lastMoveCommand = move;
				e.lastSpecialMoveCommand = special;
				e.creator = creator;
				e.direction = direction;
				robots[r].events.push_back(e);
			}

			return true;
		}
		bool pollSensors(const std::string& robotName)
		{
			int r=0;
			if(!getRobotData(robotName, r))
				return false;
			
			wok::UID callID;
			wok::sockets::Packet p;
			p.createHeader(11);
			p.packInt(robots[r].ID.get());
			p.packInt(callID.get());
			p.packChar(PACKET_VERSION);
			client.sendPacket(p);
			
			if(!waitForResponse(p, 11))
				return false;
			
			if(p.unpackInt() != callID.get())
				return false;
			
			robots[r].rotation = p.unpackFloat();
			robots[r].y = p.unpackFloat();
			robots[r].x = p.unpackFloat();
			
			robots[r].robotCollision = false;
			robots[r].trackCollision = false;
			
			if(p.unpackChar() == 1)
				robots[r].robotCollision = true;
			if(p.unpackChar() == 1)
				robots[r].trackCollision = true;
			
			int s = 0;
			int N = p.unpackChar();
			SensorData sensor;
			for(int i=0; i<N; ++i)
			{
				if(getSensorData(robotName, p.unpackInt(), r, s))
					robots[r].sensors[s].setValue(p.unpackInt());
				else
					return false;
			}
			
			return true;
		}
		
		float getRobotXPos(const std::string& robotName)
		{
			int r = 0;
			if(!getRobotData(robotName, r))
				return 0;
			
			return robots[r].x;
		}
		float getRobotYPos(const std::string& robotName)
		{
			int r = 0;
			if(!getRobotData(robotName, r))
				return 0;
			
			return robots[r].y;
		}

		float getRobotRotation(const std::string& robotName)
		{
			int r = 0;
			if(!getRobotData(robotName, r))
				return 0;
			
			return robots[r].rotation;
		}
		
		int getSensor(const std::string& robotName, const std::string& sensorName)
		{
			int r=0;
			if(!getRobotData(robotName, r))
				return 0;
			
//			std::cout << robot.trackCollision << "\n";
			if(sensorName == "trackCollision")
				return (robots[r].trackCollision==true)?1:0;
			else if(sensorName == "robotCollision")
				return (robots[r].robotCollision==true)?1:0;
			else if(sensorName == "collision")
			{
				//std::cout << (int)robots[r].trackCollision << "\n";
				return ((robots[r].trackCollision | robots[r].robotCollision)==true)?1:0;
			}
			
			int s=0;
			if(!getSensorData(robotName, sensorName, r,s))
				return 0;
			return robots[r].sensors[s].getValue();
		}
		
	private:
		bool verifyVersion(wok::sockets::Packet& p)
		{
			if(p.unpackChar() != PACKET_VERSION)
				return false;
			return true;
		}
		bool waitForResponse(wok::sockets::Packet& p, int type)
		{
			wok::sockets::Socket* socket = client.getSocket();
			if(!socket)
				return false;
			wok::Timer timeout;
			int packetType = -1;
			timeout.start(10*1000);
			while(!timeout.done())
			{
				client.mainLoop();
				packetType = socket->grabNextPacket(p);
				if(packetType == -1)
					continue;
				else if(packetType != type)
					return false;
				else if(!verifyVersion(p))
					return false;
				else
					return true;
			}
			
			return false;
		}
		std::vector <RobotData> robots;
		
		wok::sockets::Client client;
	};
};

#endif