/*
Interpreter.h
Written by Walter O. Krawec

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef WOK_INTERPRETER_H_
#define WOK_INTERPRETER_H_

#define ALLOW_MODULES

#include <string>
#include <sstream>
#include <list>

#ifdef WIN32
#include <conio.h>
#include <Windows.h>
#endif

#include <iostream>
#include <fstream>

#include "Text.h"

namespace wok
{
namespace interp
{
enum ObjectType
{
	String=0,
	Integer,
	Double,
	UserType
};

enum
{
	InterpreterVersion=1
};

class Object;

class UserObject
{
public:
	UserObject(const std::string& _type) : type(_type) {};
	virtual ~UserObject() {};

	virtual void copy(UserObject* o)
	{
		throw std::string("Cannot copy this type.");
	}

	virtual void print(std::ostream& f)
	{
		f << type << "\n";
	};

	virtual void add(Object& A, Object& B, Object& output)
	{
		throw std::string("Cannot add this type.");
	};
	virtual void subtract(Object& A, Object& B, Object& output)
	{
		throw std::string("Cannot subtract this type.");
	};
	virtual void multiply(Object& A, Object& B, Object& output)
	{
		throw std::string("Cannot multiply this type.");
	};
	virtual void divide(Object& A, Object& B, Object& output)
	{
		throw std::string("Cannot divide this type.");
	};

	virtual bool compEqual(Object* A, Object* B)
	{
		throw std::string("Cannot compare this type.");
	};

	virtual bool compGT(Object* A, Object* B)
	{
		throw std::string("Cannot compare this type.");
	};

	virtual bool compLT(Object* A, Object* B)
	{
		throw std::string("Cannot compare this type.");
	};

	const std::string& getType() const {return type;}
protected:
	std::string type;
};

class Object
{
public:
	Object() : value(NULL) {};
	Object(const std::string& _name, const ObjectType _type, void* _value) : name(_name), type(_type), value(_value) {};
	Object(const std::string* _name, const ObjectType* _type, void* _value) : name(*_name), type(*_type), value(_value) {};

	~Object() {if(name == "(wok::interpreter)temp") cleanup();}

	const ObjectType& getType() const {return type;};
	const std::string& getName() const {return name;};

	void* getValue() {return value;};

	void clear()
	{
		value = NULL;
	}
	void cleanup()
	{
		if(value)
		{
			if(type == String)
			{
				std::string* v = (std::string*) value;
				delete v;
			}
			else if(type == Integer)
			{
				int* v = (int*) value;
				delete v;
			}
			else if(type == Double)
			{
				double* v = (double*) value;
				delete v;
			}
			else if(type == UserType)
			{
				UserObject* v = (UserObject*) value;
				delete v;
			}
			else
				throw std::string("Unrecognized type");
		}
		value = NULL;
	}

	void print(std::ostream& f, bool printName, bool printAll=false)
	{
		if(!value)
		{
			std::cout << "NULL";
			return;
		}

		if(printName)
		{
			if(type == String) f << "String";
			else if(type == Integer) f << "Integer";
			else if(type == Double) f << "Double";
			else if(type == UserType)
			{
				UserObject *o = (UserObject*)value;
				f << o->getType();
			}

			f << " " << name << " = ";
		}

		switch(type)
		{
		case String:
			f << *(std::string*)value;
			break;

		case Integer:
			f << *(int*)value;
			break;

		case Double:
			f << *(double*)value;
			break;

		case UserType:
			UserObject* o = (UserObject*)value;
			o->print(f);
			break;
		}
	}

	void setValue(void* v) {value = v;}
	void setName(const std::string& n) {name = n;}
	void setType(ObjectType t) {type = t;}

	void set(Object& input)
	{
		if(!input.value || input.type != type)
			return;

		switch(type)
		{
		case String:
			*(std::string*)value = *(std::string*)input.value;
			break;

		case Integer:
			*(int*)value = *(int*)input.value;
			break;

		case Double:
			*(double*)value = *(double*)input.value;
			break;

		case UserType:
			if(((UserObject*)value)->getType() != ((UserObject*)input.value)->getType())
				return;
			((UserObject*)value)->copy((UserObject*)input.value);
			//*(UserObject*)value = *(UserObject*)input.value;
			break;

		default:
			throw std::string("Object::set unkown type");
		}
	}

	bool compEqual (Object& other);
	bool compLT    (Object& other);
	bool compGT    (Object& other);

	void add(Object& A, Object& B, Object& output);
	void subtract(Object& A, Object& B, Object& output);
	void multiply(Object& A, Object& B, Object& output);
	void divide(Object& A, Object& B, Object& output);

	void cast(Object& input, Object& output, ObjectType newType);

	void saveToFile(std::ofstream& f);
private:
	std::string name;
	ObjectType type;
	void* value;
};

class ObjectHelper
{
public:
	static void newString(const std::string& value, Object& output)
	{
		if(output.getName() == "(wok::interpreter)temp")
			output.cleanup();
		output.setValue(new std::string(value));
		output.setName("(wok::interpreter)temp");
		output.setType(String);
	}

	static void newInteger(int value, Object& output)
	{
		if(output.getName() == "(wok::interpreter)temp")
			output.cleanup();
		int *i = new int;
		*i = value;
		output.setValue(i);
		output.setName("(wok::interpreter)temp");
		output.setType(Integer);
	}

	static void newDouble(double value, Object& output)
	{
		if(output.getName() == "(wok::interpreter)temp")
			output.cleanup();
		double *i = new double;
		*i = value;
		output.setValue(i);
		output.setName("(wok::interpreter)temp");
		output.setType(Double);
	}

	static void newUserObject(UserObject* value, Object& output)
	{
		if(output.getName() == "(wok::interpreter)temp")
			output.cleanup();

		output.setValue(value);
		output.setName("(wok::interpreter)temp");
		output.setType(UserType);
	}

	static std::string* getString(Object& input)
	{
		if(!input.getValue() || input.getType() != String)
			throw std::string("getString error - type is NULL or not string");

		return (std::string*)input.getValue();
	}
	static int* getInteger(Object& input)
	{
		if(!input.getValue() || input.getType() != Integer)
		{
			throw std::string("getInteger error - type is NULL or not int");
		}

		return (int*)input.getValue();
	}
	static double* getDouble(Object& input)
	{
		if(!input.getValue() || input.getType() != Double)
		{
			throw std::string("getDouble error - type is NULL or not double");
		}

		return (double*)input.getValue();
	}
	static UserObject* getUserObject(Object& input)
	{
		if(!input.getValue() || input.getType() != UserType)
			throw std::string("getUserObject error - type is NULL or not UserType");

		return (UserObject*)input.getValue();
	}
};

class Store
{
public:
	~Store() {cleanup();}
	void cleanup()
	{
		if(!store.empty())
		{
			std::list <Pair>::iterator Iter;
			for(Iter = store.begin(); Iter != store.end(); ++Iter)
			{
				if(!Iter->isCopy)
					Iter->object.cleanup();
			}
		}
		store.clear();
	}
	void set(const std::string& name, Object& input)
	{
		if(store.empty()) return;
		std::list <Pair>::iterator Iter;
		for(Iter = store.begin(); Iter != store.end(); ++Iter)
		{
			if(Iter->object.getName() == name)
				Iter->object.set(input);
		}
	}
	void get(const std::string& name, Object& output)
	{
		if(store.empty()) return;
		std::list <Pair>::iterator Iter;
		for(Iter = store.begin(); Iter != store.end(); ++Iter)
		{
			if(Iter->object.getName() == name)
				output = Iter->object;
		}
	}
	void addStore(Object& v)
	{
		store.push_back(Pair(v, false));
	}
	void removeStore(std::string& name)
	{
		if(store.empty()) return;
		std::list <Pair> ::iterator Iter;
		for(Iter = store.begin(); Iter != store.end(); ++Iter)
		{
			if(Iter->object.getName() == name)
			{
				if(Iter->isCopy)
					throw std::string("Cannot delete variable");
				Iter->object.cleanup();
				store.erase(Iter);
				return;
			}
		}
	}
	void clear()
	{
		store.clear();
	}

	void copy(Store& newStore)
	{
		newStore.cleanup();

		if(!store.empty())
		{
			std::list <Pair>::iterator Iter;
			for(Iter = store.begin(); Iter != store.end(); ++Iter)
			{
				newStore.store.push_back(Pair(Iter->object, true));
			}
		}
	}

	void print()
	{
		std::cout << "State List:\n";
		if(store.empty()) return;

		std::list <Pair>::iterator Iter;
		for(Iter = store.begin(); Iter != store.end(); ++Iter)
		{
			std::cout << "<" << Iter->object.getType() << ", " << Iter->object.getName() << ">\n";
		}
	}

	void saveToFile(std::ofstream& f);
	bool loadFromFile(std::ofstream& f);
private:
	struct Pair
	{
		Pair() : isCopy(false) {};
		Pair(Object& o, bool _iscopy) : object(o), isCopy(_iscopy) {};

		Object object;
		bool isCopy;
	};
	std::list <Pair> store;
};

class Interpreter;

class CommandLibrary
{
public:
	virtual ~CommandLibrary() {};
	virtual bool evaluate(wok::Text& text, Store& store, Object& output, Interpreter* interpreter, std::vector <CommandLibrary*> &library, bool globalMode)=0;
	virtual bool newUserType(const std::string& type, const std::string& name, Store* store)=0;
	virtual int requiredVersion()=0;
private:
};

struct Module
{
	Module() : library(NULL) {};
	Module(const std::string& _name, CommandLibrary* lib) : library(lib), name(_name) {};
	CommandLibrary* library;
	std::string name;
};

class Interpreter
{
	Interpreter(const Interpreter&) {};	// can't copy
public:
	Interpreter() {};
	virtual ~Interpreter()
	{
		for(int i=0; i<(int)modules.size(); ++i)
		{
			if(modules[i].library)
				delete modules[i].library;
		}

		modules.clear();

#ifdef WIN32
		for(int i=0; i<(int)moduleHandles.size(); ++i)
		{
			FreeLibrary(moduleHandles[i]);
		}
		moduleHandles.clear();
#endif
	}

	virtual void evaluate(const std::string& line, Store& store, Object& output, std::vector <CommandLibrary*> &library, bool globalMode=false, bool copyStore=true);

	void preProcess(std::string& text)
	{
		if(text.size() == 0) return;
		std::stringstream str;
		bool nonSpace = false;
		int numSpaces = 0;
		unsigned int endIndex = text.size();
		for(unsigned int i=text.size()-1; i>=0; --i)
		{
			if(text[i] != ' ')
			{
				endIndex = i+1;
				break;
			}
		}

		bool skipSpace = false;
		for(unsigned int i=0; i<endIndex; ++i)
		{
			if(text[i] == '(' && nonSpace && numSpaces == 0)
				str << ' ';
			if(text[i] != ' ')
				nonSpace = true;

			if(text[i] != ' ' && text[i] != '\t')
				numSpaces = 0;

			if(nonSpace)
			{
				if((text[i] == ' ' && !skipSpace) || text[i] != ' ')
				str << text[i];
			}
			//if(text[i] == ')' && nonSpace && numSpaces == 0)
			//	str << ' ';
			if(text[i] == ' ')
			{
				++numSpaces;
				nonSpace = false;
			}
			else if(text[i] == '(')
				skipSpace = true;
			else
				skipSpace = false;
		}
		text = std::string(str.str());
	}

	void shell(const char* shellName, std::vector <CommandLibrary*> &library)
	{
		wok::Text text;
		text.addDelimiter(' ');

		Object output;
		int bufferSize = 2048;
		char* buffer = new char[bufferSize];

		while(1)
		{
			if(shellName)
				std::cout << shellName << " > ";
			else
				std::cout << "> ";

			std::cin.getline(buffer, bufferSize-2);
			if(buffer[0] == 'e' && buffer[1] == 'x' && buffer[2] == 'i' && buffer[3] == 't' && buffer[4] == 0)
				break;
			if(buffer[0] == '!')
			{
				text.parse(buffer);
				if(text[0] == "!exit")
					break;
				else if(text[0] == "!store")
					printGlobal();
				else if(text[0] == "!clear")
					clearGlobalStore();
				else if(text[0] == "!path")
				{
					for(int i=0; i<(int)path.size(); ++i)
						std::cout << path[i] << "\n";
				}
			}
			else
			{
			try {
				std::string line(buffer);
				preProcess(line);
				evaluate(line, getGlobalStore(), output, library, false);
			}
			catch(std::string& e)
			{
				std::cout << "Error: " << e << "\n";
			}

			output.print(std::cout, true);
			}
			std::cout << "\n\n";
		}
	}

	Store& getGlobalStore() {return globalStore;}
	void clearGlobalStore()
	{
		globalStore.cleanup();
	}
	void printGlobal()
	{
		globalStore.print();
	}

	void addToPath(const std::string& p)
	{
		path.push_back(p);
	}

private:
	Store globalStore;

	std::vector <Module> modules;

	std::vector <std::string> path;
#ifdef WIN32
	std::vector <HMODULE> moduleHandles;
#endif
};

#ifdef WIN32
typedef CommandLibrary* (__cdecl *moduleCallBack)();
#endif

}
};

#endif