/*
Sockets.h
Written by Walter O. Krawec

Copyright (c) 2012 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*
This is a cross-platform socket library (tested on Windows and Mac OS)
*/

#ifndef _SOCKETS_H_
#define _SOCKETS_H_

#define uint32 u_long

#ifndef WIN32

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <sys/fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <iostream>

#else
#include <winsock2.h>
#endif

#include "timer.h"
#include "UID.h"
#include <string.h>
#include <sstream>
#include <iostream>
#include <vector>
#include <list>

namespace wok
{
namespace sockets
{	
	enum
	{
		HEADER_SIZE=5,
		MAX_CLIENTS=10,
		MAX_PACKET_SIZE=500,
		MAX_RECV_BUFFER=1024
	};
	
	class Socket;
	class MultiPacket;
	// When sending a packet send: PACKET_ID (BYTE), PACKET_SIZE (BYTE), PACKET_DATA
	// PACKET_SIZE should count the header size (i.e. PACKET_SIZE = sizeof(PACKET_DATA)+2
	class Packet
	{
	public:
		friend class Socket;
		Packet() : bufferIndex(0) {clear();};
		
		void clear()
		{
			packetData[0] = -1;
			bufferIndex = 0;
		};
		
		bool packChar(char data);
		bool packInt(int data);
		bool packFloat(float data);
		bool packChars(char* data, int dataSize);
		bool packString(const std::string& str);
		
		char unpackChar();
		int unpackInt();
		float unpackFloat();
		void unpackString(std::string& str);
		
		void debugPrintPacket();
		
		char* getPacketData() {return packetData;};
		int packetSize() {return bufferIndex;};
		
		void createHeader(char packetID)
		{
			clear();
			packChar(packetID);
			packInt(5);
		}
		void updateHeader();
		char getPacketID(){return packetData[0];};
	private:
		char packetData[MAX_PACKET_SIZE];
		int bufferIndex;

		char getBufferElement(int index)
		{
			if(index >= 0 && index < MAX_PACKET_SIZE)
				return packetData[index];
			return 0;
		}
		void setBufferElement(int index, char c)
		{
			if(index >= 0 && index < MAX_PACKET_SIZE)
				packetData[index] = c;
		}
		void setBufferIndex(int index)
		{
			if(index >= 0 && index < MAX_PACKET_SIZE)
				bufferIndex = index;
		}
	};

	class MultiPacket
	{
	public:
		friend class Socket;
		MultiPacket() : sendPosition(0), type(0), complete(true) {};

		void addPacket(Packet& p)
		{
			p.updateHeader();
			packetList.push_back(p);
		};

		void getPacket(int index, Packet& p)
		{
			if(index >= 0 && index < (int)packetList.size())
				p = packetList[index];
		};

		void getLastPacket(int offset, Packet& p)
		{
			getPacket((int)packetList.size()-1-offset, p);
		};

		int getPacketListSize()
		{
			return (int)packetList.size();
		}

		void newID()
		{
			ID = wok::UID();
		}
	private:
		std::vector <Packet> packetList;
		int sendPosition, type;
		bool complete;
		wok::UID ID;
	};
	
	class Socket
	{
	public:
		Socket()
		{
			numCompleteMP = 0;
			bufferIndex = 0;
			socketID = -1;
		}
		~Socket() {shutdownSocket();};
		void shutdownSocket()
		{
			bufferIndex = 0;
			if(socketID != -1)
			{
				std::cout << "closing socket.\n";
#ifndef WIN32
				shutdown(socketID, SHUT_RDWR);
				close(socketID);
#else
				shutdown(socketID, SD_SEND);
				closesocket(socketID);
#endif
			}
			socketID = -1;
		}
		
		bool setupSocket()
		{
			shutdownSocket();
			socketID = socket(AF_INET, SOCK_STREAM, 0);
			if(socketID == -1)
				return false;
			return true;
		};
		void setNonBlocking()
		{
#ifndef WIN32
			if(socketID == -1)
				return;
			int flags = fcntl(socketID, F_GETFL, 0);
			if(flags < 0) return;
			
			fcntl(socketID, F_SETFL, flags | O_NONBLOCK);
			
			signal(SIGPIPE, SIG_IGN);
#endif
		};

		bool shutdownWinSock();
		
		bool ping();
		void test()
		{
			std::cout << "inbox = " << multiPacketInbox.size() << "\n";
			std::cout << "outbox = " << multiPacketOutbox.size() << "\n";
		};
		
		bool sendPacket(Packet& p);
		void checkForData();	// run this as often as possible.
		
#ifndef WIN32
		int getSocketID() {return socketID;};
		void setSocketID(int ID) {socketID = ID;};
#else
		SOCKET getSocketID() {return socketID;};
		void setSocketID(SOCKET ID) {socketID = ID;};
#endif
		
		int recvBufferSize() const {return bufferIndex; };
		
//		PacketHeader nextPacketInfo();
		
		int grabNextPacket(Packet& packet); // return -1 if no packet ready otherwise returns packetID
		int nextPacketType();
		
		bool dataWaiting()
		{
			if(bufferIndex != 0 && socketID != -1)
				return true;
			return false;
		};

		void sendMultiPacket(MultiPacket& mp);
		bool recvMultiPacket(MultiPacket& mp);
		void processPacket(Packet& p);
		void mainLoop();
	private:
	//	bool initWinSock();
		void sendMultiPacket(std::list <MultiPacket>::iterator& Iter);
#ifndef WIN32
		int socketID;
#else
		SOCKET socketID;
#endif
		char recvBuffer[MAX_RECV_BUFFER];
		int bufferIndex;

		std::list <MultiPacket> multiPacketInbox, multiPacketOutbox;
		int numCompleteMP;
	};

	class WinSockManager
	{
		WinSockManager()
		{
			winSockCount = 0;
		}
	public:
		~WinSockManager()
		{
			if(winSockCount > 0)
				shutDownWinSock();
		}
		static WinSockManager& instance()
		{
			static WinSockManager manager;
			return manager;
		}

		bool initWinSock()
		{
			++winSockCount;
			if(winSockCount > 1)
				return true;
#ifdef WIN32
			WSADATA WSVersion;

			int error = WSAStartup(0x0202, &WSVersion);

			if(error)
				return false;

			if(WSVersion.wVersion != 0x0202)
			{
				WSACleanup();
				return false;
			}
#endif
			return true;
		}
	private:
		void shutDownWinSock()
		{
			WSACleanup();
		}
		int winSockCount;
	};
	
	class Host
	{
	public:
		Host() : timeToPing(2000), keepLog(false), printLogToScreen(true) {};
		~Host(){shutdownHost();};
		void shutdownHost()
		{
			mainSocket.shutdownSocket();
			for(int i=0; i<MAX_CLIENTS; ++i)
				clientSockets[i].shutdownSocket();
		}
		bool setupHost(unsigned short portNumber, int _timeToPing=2000);
		bool lookForConnection();
		
		void mainLoop();
		
		int recvBufferSize(int clientIndex) const {return clientSockets[clientIndex].recvBufferSize();};
		
		bool dataWaiting(int clientID);
		Socket* getClientSocket(int clientID);
		bool sendPacket(Packet& p, int clientID);	// clientID=-1 to send to all clients
		
		void disconnectClient(int clientID);

		Socket* getMainSocket() {return &mainSocket;};
	private:
		Socket mainSocket;
		Socket clientSockets[MAX_CLIENTS];
		wok::Timer pingClients;
		int timeToPing;
		
		bool keepLog, printLogToScreen;
	};
	
	class Client
	{
	public:
		void disconnect()
		{
			mainSocket.shutdownSocket();
		}
		bool connectToHost(const char* IPAddress, unsigned short portNumber);
		bool sendPacket(Packet& p)
		{
			return mainSocket.sendPacket(p);
		}
		
		Socket* getSocket() {return &mainSocket;};
		
		void mainLoop();
		
		bool isConnected()
		{
			if(mainSocket.getSocketID() == -1)
				return false;
			return true;
		}
	private:
		Socket mainSocket;
	};
};
};

#endif